# AA RSS to Discord
from aa_rss_to_discord.models import RssFeeds
from aa_rss_to_discord.tests import BaseTestCase


class TestRssFeedManager(BaseTestCase):
    """
    Test the RSS Feed Manager
    """

    def test_returns_only_enabled_feeds(self):
        RssFeeds.objects.create(name="Feed 1", enabled=True)
        RssFeeds.objects.create(name="Feed 2", enabled=False)
        RssFeeds.objects.create(name="Feed 3", enabled=True)

        result = RssFeeds.objects.select_enabled()

        self.assertEqual(result.count(), 2)
        self.assertTrue(all(feed.enabled for feed in result))

    def test_handles_no_enabled_feeds_gracefully(self):
        RssFeeds.objects.create(name="Feed 1", enabled=False)
        RssFeeds.objects.create(name="Feed 2", enabled=False)

        result = RssFeeds.objects.select_enabled()

        self.assertEqual(result.count(), 0)

    def test_handles_empty_database_gracefully(self):
        result = RssFeeds.objects.select_enabled()

        self.assertEqual(result.count(), 0)



# Standard Library
from unittest import mock

# AA RSS to Discord
from aa_rss_to_discord.auth_hooks import register_cogs
from aa_rss_to_discord.tests import BaseTestCase


class TestAuthHooks(BaseTestCase):
    """
    Test the auth hooks
    """

    def test_registers_discord_cogs_hook_correctly(self):
        """
        Tests that the discord cogs hook is registered correctly.

        :return:
        :rtype:
        """

        result = register_cogs()

        self.assertIn("aa_rss_to_discord.discordbot.cogs.rss", result)

    def test_returns_list_of_cogs(self):
        """
        Tests that the register_cogs function returns a list.

        :return:
        :rtype:
        """

        result = register_cogs()

        self.assertIsInstance(result, list)

    def test_handles_empty_hook_registration_gracefully(self):
        """
        Tests that the register_cogs function handles an empty hook registration gracefully.

        :return:
        :rtype:
        """

        with mock.patch("allianceauth.hooks.register", return_value=None):
            result = register_cogs()

            self.assertIsNotNone(result)



"""
Test cases for the fetch_rss task
"""

# Standard Library
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

# AA RSS to Discord
from aa_rss_to_discord.constants import USER_AGENT
from aa_rss_to_discord.models import RssFeeds
from aa_rss_to_discord.tasks import _process_feed, fetch_rss
from aa_rss_to_discord.tests import BaseTestCase


class TestFetchRss(BaseTestCase):
    """
    Test cases for the fetch_rss task
    """

    def test_fetch_rss_logs_message_when_no_rss_feeds_found(self):
        """
        Test that fetch_rss logs a message when no RSS feeds are found.

        :return:
        :rtype:
        """

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_enabled"
            ) as mock_select_enabled,
            patch("aa_rss_to_discord.tasks.logger.debug") as mock_logger_debug,
        ):
            mock_select_enabled.return_value = []

            fetch_rss()

            mock_logger_debug.assert_called_once_with("No RSS feeds found to parse.")

    def test_fetch_rss_dispatches_tasks_for_enabled_feeds(self):
        """
        Test that fetch_rss dispatches tasks for enabled RSS feeds.

        :return:
        :rtype:
        """

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_enabled"
            ) as mock_select_enabled,
            patch("aa_rss_to_discord.tasks.group") as mock_group,
        ):
            mock_rss_feed_1 = MagicMock(id=1)
            mock_rss_feed_2 = MagicMock(id=2)
            mock_select_enabled.return_value = [mock_rss_feed_1, mock_rss_feed_2]

            fetch_rss()

            mock_group.assert_called_once()
            mock_group.return_value.apply_async.assert_called_once()
            dispatched_tasks = mock_group.call_args[0][0]
            dispatched_list = list(dispatched_tasks)
            self.assertEqual(len(dispatched_list), 2)
            self.assertEqual(dispatched_list[0].args[0], 1)
            self.assertEqual(dispatched_list[1].args[0], 2)

    def test_handles_empty_rss_feed_ids_gracefully(self):
        """
        Test that fetch_rss handles an empty list of RSS feed IDs gracefully.

        :return:
        :rtype:
        """

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_enabled"
            ) as mock_select_enabled,
            patch("aa_rss_to_discord.tasks.group") as mock_group,
            patch("aa_rss_to_discord.tasks.logger.debug") as mock_logger_debug,
        ):
            mock_select_enabled.return_value = iter([])

            fetch_rss()

            mock_group.assert_not_called()
            mock_logger_debug.assert_called_once_with("No RSS feed ids to dispatch.")


class TestHelperProcessRssFeed(BaseTestCase):
    """
    Test cases for the _process_feed task
    """

    def test_handles_missing_rss_feed_gracefully(self):
        """
        Test that _process_feed handles a missing RSS feed gracefully.

        :return:
        :rtype:
        """

        with patch(
            "aa_rss_to_discord.tasks.RssFeeds.objects.select_related"
        ) as mock_select_related:
            mock_select_related.return_value.get.side_effect = RssFeeds.DoesNotExist

            _process_feed(1)

            mock_select_related.return_value.get.assert_called_once_with(id=1)

    def test_logs_and_skips_when_no_entries_in_feed(self):
        """
        Test that _process_feed logs and skips processing when no entries are found in the feed.

        :return:
        :rtype:
        """

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_related"
            ) as mock_select_related,
            patch(
                "aa_rss_to_discord.tasks.feedparser.parse", return_value={}
            ) as mock_parse,
        ):
            mock_rss_feed = MagicMock(url="http://example.com", name="No Entries Feed")
            mock_select_related.return_value.get.return_value = mock_rss_feed

            _process_feed(1)

            mock_parse.assert_called_once_with(mock_rss_feed.url, agent=USER_AGENT)

    def test_posts_new_entry_to_discord(self):
        """
        Test that _process_feed posts a new entry to Discord when it's not a duplicate.

        :return:
        :rtype:
        """

        entry = {"title": "New Entry", "link": "http://example.com", "id": "123"}

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_related"
            ) as mock_select_related,
            patch(
                "aa_rss_to_discord.tasks.feedparser.parse",
                return_value=SimpleNamespace(entries=[entry]),
            ) as mock_parse,
            patch("aa_rss_to_discord.tasks.LastItem.objects.filter") as mock_filter,
            patch(
                "aa_rss_to_discord.tasks.LastItem.objects.update_or_create"
            ) as mock_update_or_create,
            patch("aa_rss_to_discord.tasks.send_message") as mock_send_message,
        ):
            mock_rss_feed = MagicMock(
                discord_channel=MagicMock(channel=12345),
                name="Test Feed",
                url="http://example.com",
            )
            mock_select_related.return_value.get.return_value = mock_rss_feed
            mock_filter.return_value.first.return_value = None
            mock_update_or_create.return_value = (MagicMock(), True)

            _process_feed(1)

            mock_parse.assert_called_once_with(mock_rss_feed.url, agent=USER_AGENT)
            mock_update_or_create.assert_called_once()
            self.assertTrue(
                mock_send_message.called,
                f"send_message not called. calls: {mock_send_message.call_args_list}",
            )

            call_args, call_kwargs = mock_send_message.call_args
            called_channel = (
                call_kwargs.get("channel_id")
                if "channel_id" in call_kwargs
                else (call_args[0] if len(call_args) > 0 else None)
            )
            called_message = (
                call_kwargs.get("message")
                if "message" in call_kwargs
                else (
                    call_args[1]
                    if len(call_args) > 1
                    else (call_args[0] if len(call_args) == 1 else None)
                )
            )

            self.assertEqual(called_channel, 12345)
            self.assertIsNotNone(called_message)
            self.assertIn(mock_rss_feed.url, called_message)

    def test_skips_posting_duplicate_entry(self):
        """
        Test that _process_feed skips posting a duplicate entry to Discord.

        :return:
        :rtype:
        """

        entry = {"title": "Duplicate Entry", "link": "http://example.com", "id": "123"}

        with (
            patch(
                "aa_rss_to_discord.tasks.RssFeeds.objects.select_related"
            ) as mock_select_related,
            patch(
                "aa_rss_to_discord.tasks.feedparser.parse",
                return_value=SimpleNamespace(entries=[entry]),
            ),
            patch("aa_rss_to_discord.tasks.LastItem.objects.filter") as mock_filter,
            patch("aa_rss_to_discord.tasks.send_message") as mock_send_message,
        ):
            mock_rss_feed = MagicMock(
                discord_channel=MagicMock(channel=12345),
                name="Dup Feed",
                url="http://example.com",
            )
            mock_select_related.return_value.get.return_value = mock_rss_feed
            mock_existing = MagicMock(
                rss_item_time=None,
                rss_item_title="Duplicate Entry",
                rss_item_link="http://example.com",
                rss_item_guid="123",
            )
            mock_filter.return_value.first.return_value = mock_existing

            _process_feed(1)

            mock_send_message.assert_not_called()



"""
AA RSS To Discord Tasks
"""

# Standard Library
import re

# Third Party
import feedparser
from aadiscordbot.tasks import send_message
from celery import group, shared_task

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# AA RSS to Discord
from aa_rss_to_discord import __title__
from aa_rss_to_discord.constants import USER_AGENT
from aa_rss_to_discord.models import LastItem, RssFeeds
from aa_rss_to_discord.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)


def remove_emoji(string: str) -> str:
    """
    Removing these dumb as fuck emojis from the title string.
    Like honestly, who in the hell needs that shit?

    :param string: Input string
    :type string: str
    :return: String without emojis
    :rtype: str
    """

    emoji_pattern = re.compile(
        pattern="["
        "\U0001f600-\U0001f64f"  # Emoticons
        "\U0001f300-\U0001f5ff"  # Symbols & pictographs
        "\U0001f680-\U0001f6ff"  # Transport & map symbols
        "\U0001f1e0-\U0001f1ff"  # Flags (iOS)
        "\U00002500-\U00002bef"  # Chinese char
        "\U00002702-\U000027b0"
        "\U00002702-\U000027b0"
        "\U000024c2-\U0001f251"
        "\U0001f926-\U0001f937"
        "\U00010000-\U0010ffff"
        "\u2640-\u2642"
        "\u2600-\u2b55"
        "\u200d"
        "\u23cf"
        "\u23e9"
        "\u231a"
        "\ufe0f"  # Dingbats
        "\u3030"
        "]+",
        flags=re.UNICODE,
    )

    return emoji_pattern.sub(repl=r"", string=string)


@shared_task(**{"base": QueueOnce})
def fetch_rss() -> None:
    """
    Fetch all enabled RSS feeds and dispatch processing tasks.

    :return: None
    :rtype: None
    """

    rss_feeds = RssFeeds.objects.select_enabled()
    if not rss_feeds:
        logger.debug("No RSS feeds found to parse.")
        return

    feed_ids = [f.id for f in rss_feeds]
    if not feed_ids:
        logger.debug("No RSS feed ids to dispatch.")
        return

    group(_process_feed.s(fid) for fid in feed_ids).apply_async()


@shared_task
def _process_feed(rss_feed_id: int) -> None:
    """
    Process a single RSS feed by fetching its latest entry and posting to Discord if it's new.

    :param rss_feed_id: ID of the RSS feed to process
    :type rss_feed_id: int
    :return: None
    :rtype: None
    """

    try:
        rss_feed = RssFeeds.objects.select_related("discord_channel").get(
            id=rss_feed_id
        )
    except RssFeeds.DoesNotExist:
        logger.debug("RSS feed %s not found", rss_feed_id)
        return

    logger.info(f'Fetching RSS Feed "{rss_feed.name}"')
    feed = feedparser.parse(rss_feed.url, agent=USER_AGENT)

    latest_entry = next(iter(getattr(feed, "entries", [])), None)
    if not latest_entry:
        logger.debug(f'No entries found for feed "{rss_feed.name}".')
        return

    feed_entry_title = remove_emoji(latest_entry.get("title", "No title"))
    feed_entry_link = latest_entry.get("link")
    feed_entry_time = latest_entry.get("published", latest_entry.get("updated"))
    feed_entry_guid = latest_entry.get("id")

    last_item = LastItem.objects.filter(rss_feed=rss_feed).first()
    if last_item and (
        last_item.rss_item_time,
        last_item.rss_item_title,
        last_item.rss_item_link,
        last_item.rss_item_guid,
    ) == (feed_entry_time, feed_entry_title, feed_entry_link, feed_entry_guid):
        logger.debug(
            'News item "%s" for RSS Feed "%s" has already been posted to your Discord',
            feed_entry_title,
            rss_feed.name,
        )
        return

    if not last_item:
        logger.debug("This seems to be a completely new RSS feed: %s", rss_feed.name)

    logger.info(
        'New entry found for RSS feed "%s", posting to Discord channel %s',
        rss_feed.name,
        rss_feed.discord_channel,
    )

    LastItem.objects.update_or_create(
        rss_feed=rss_feed,
        defaults={
            "rss_item_time": feed_entry_time,
            "rss_item_title": feed_entry_title,
            "rss_item_link": feed_entry_link,
            "rss_item_guid": feed_entry_guid,
        },
    )

    send_message(
        channel_id=rss_feed.discord_channel.channel,
        message=f"**{rss_feed.name}**\n{feed_entry_link}",
    )



# Third Party
from aadiscordbot.models import Channels, Servers

# AA RSS to Discord
from aa_rss_to_discord.models import LastItem, RssFeeds
from aa_rss_to_discord.tests import BaseTestCase


class TestRssFeeds(BaseTestCase):
    """
    Test the RssFeeds model.
    """

    def test_saves_and_retrieves_rss_feed_correctly(self):
        """
        Tests that an RSS feed can be saved and retrieved correctly.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(
            name="Feed 1", url="http://example.com/rss", enabled=True
        )
        retrieved_feed = RssFeeds.objects.get(id=feed.id)

        self.assertEqual(retrieved_feed.name, "Feed 1")
        self.assertEqual(retrieved_feed.url, "http://example.com/rss")
        self.assertTrue(retrieved_feed.enabled)

    def test_handles_null_discord_channel_gracefully(self):
        """
        Tests that the model handles a null discord_channel gracefully.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(
            name="Feed 2", url="http://example.com/rss", discord_channel=None
        )

        self.assertIsNone(feed.discord_channel)

    def test_string_representation_includes_name_and_channel(self):
        """
        Tests that the string representation of the RssFeeds model

        :return:
        :rtype:
        """
        server = Servers.objects.create(server=1, name="Test Server")
        channel = Channels.objects.create(name="General", server=server, channel=1)
        feed = RssFeeds.objects.create(name="Feed 3", discord_channel=channel)

        self.assertEqual(
            str(feed), 'RSS Feed "Feed 3" for channel "General" On "Test Server"'
        )

    def test_allows_disabling_rss_feed(self):
        """
        Tests that an RSS feed can be disabled.

        :return:
        :rtype:
        """
        feed = RssFeeds.objects.create(name="Feed 4", enabled=True)
        feed.enabled = False
        feed.save()

        updated_feed = RssFeeds.objects.get(id=feed.id)

        self.assertFalse(updated_feed.enabled)

    def test_handles_empty_name_and_url_gracefully(self):
        """
        Tests that the model handles empty name and url fields gracefully.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(name="", url="")

        self.assertEqual(feed.name, "")
        self.assertEqual(feed.url, "")


class TestLastItem(BaseTestCase):
    """
    Test the LastItem model.
    """

    def test_string_representation_includes_item_title(self):
        """
        Tests that the string representation of the LastItem model includes the item title.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(name="Feed 1", url="http://example.com/rss")
        item = LastItem.objects.create(
            rss_feed=feed,
            rss_item_title="Sample Item",
            rss_item_link="http://example.com/item",
        )

        self.assertEqual(str(item), 'RSS Entry "Sample Item"')

    def test_handles_empty_item_title_gracefully(self):
        """
        Tests that the model handles an empty item title gracefully.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(name="Feed 2", url="http://example.com/rss")
        item = LastItem.objects.create(
            rss_feed=feed,
            rss_item_title="",
            rss_item_link="http://example.com/item",
        )

        self.assertEqual(str(item), 'RSS Entry ""')

    def test_saves_and_retrieves_last_item_correctly(self):
        """
        Tests that a LastItem can be saved and retrieved correctly.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(name="Feed 3", url="http://example.com/rss")
        item = LastItem.objects.create(
            rss_feed=feed,
            rss_item_time="2023-10-01T12:00:00Z",
            rss_item_title="Item Title",
            rss_item_link="http://example.com/item",
            rss_item_guid="12345",
        )
        retrieved_item = LastItem.objects.get(id=item.id)

        self.assertEqual(retrieved_item.rss_item_time, "2023-10-01T12:00:00Z")
        self.assertEqual(retrieved_item.rss_item_title, "Item Title")
        self.assertEqual(retrieved_item.rss_item_link, "http://example.com/item")
        self.assertEqual(retrieved_item.rss_item_guid, "12345")

    def test_handles_null_guid_gracefully(self):
        """
        Tests that the model handles a null guid gracefully.

        :return:
        :rtype:
        """

        feed = RssFeeds.objects.create(name="Feed 4", url="http://example.com/rss")
        item = LastItem.objects.create(
            rss_feed=feed,
            rss_item_guid="",
        )

        self.assertEqual(item.rss_item_guid, "")
